=begin rdoc

Модуль описыват класс Storage, который содержит
все действия(методы) производимые над хранилищем.

Под хранилишем понимается сосокупность базы данных
(прежде всего ссылок на файлы в таблице EVENTS)
и файлов в фаловой системе.

Именно эти действия вызываются из программы avreg-unlink.
=end

class Storage
   SCRIPT_DIR = '/etc/avreg/scripts/unlink'
   SCRIPT_REGX = /^[A-Za-z0-9][A-Za-z0-9_\-\.]+$/

   attr_reader :opts
   def initialize(opts)
      @opts = opts
      @fs = SorageFS.new(@opts)
      # connect to db
      begin
         @db = AVReg_DB.new(opts)
      rescue
         @fs.close!
         raise
      end

      # /etc/avreg/scripts/unlink/{before,after}/SCRIPT_REGX
      @before_scripts = get_scripts('before')
      @after_scripts = get_scripts('after')
      @env_hash = opts.to_env
   end

   def close!
      fetched_rows = @db.fetched
      @db.close! if not @db.nil?

      if fetched_rows > 0
         exec_scripts('after')
      end

      @fs.close! if not @fs.nil?
      # free resource ?
      remove_instance_variable(:@db)
      remove_instance_variable(:@fs)
   end

   def repair_db
      $errnct += @db.repair if @opts['repair']
   end

   def optimize_db
      $errnct += @db.optimize if @opts['optimize']
   end

   def manual_deletion?
      @opts['to-date'] || @opts['from-date']
   end

   def free_space_based_deletion?
      not manual_deletion?
   end

   def unlink
      #########
      # stage 2 : delete
      $log.notice _('Start AVReg media archive cleaner on') + ' "' + @db.db_string + '@' + @opts['storage-dir'] + '"'
      @fs.PrintBeforeStat.each { |line| $log.info(line) }

      if not @opts['del']
         $log.notice _("skip deleting by `--no-del' options")
         return
      end

      if free_space_based_deletion? and @fs.start_free_space > @opts['start-space']
         $log.notice(sprintf(_("free space") + ' (%s) > "start-space" options (%s) => '+ _('skip deleting') + ".\n",
                             @fs.start_free_space.to_hfs, @opts['start-space'].to_hfs))
         return
      end


      exec_scripts('before')

      # удаляем
      total_deleted_record = 0
      # читаем порциями, удаляем пока только файлы
      @db.select_olderst do |row|
         # XXX return value for yield
         if free_space_based_deletion?
            @fs.unlink!(row) or $quit
         else
            # do not check free space
            @fs.unlink!(row)
            $quit
         end
      end

      # теперь удаляем из базы всё что начитали
      if @db.fetched > 0
         if $verbose >= 2
            $log.notice _(" total fetched %d rows, try delete it") % @db.fetched
         end
         total_deleted_record = @db.del_fetched!
         if @db.fetched != total_deleted_record
            $log.emerg _("DB records delete error => need %{fetched}, really deleted %{deleted}") %
            { :fetched => @db.fetched, :deleted => total_deleted_record }
         end
         if $verbose >= 1
            $log.notice _(" deleted %d rows from EVENTS table") % total_deleted_record
         end
      end

      @fs.unlink_empty_dirs!

      @fs.PrintAfterStat.each { |line| if @fs.fetched_files > 0 then $log.notice(line) else $log.info(line) end }

      @fs.PrintWarn
   end

   protected
   def get_scripts(stage)
      script_dir = SCRIPT_DIR + '/' + stage
      d = Dir.new(script_dir)
      ret = []
      d.each  { |x|
         stat = File.stat(script_dir + '/' + x)
         if stat.file? and stat.executable_real?
            if SCRIPT_REGX =~ x
               if  stat.uid == 0 and stat.gid == 0 and stat.mode == 0100755
                  ret << x
               else
                  $log.warning _('ignore "%{file}" script by wrong permissions, should be root:root 0755') %
                     {:file => stage + '/' + x}
               end
            end
         end
      }
      ret.sort
   rescue
      []
   end

   def exec_scripts(stage)
      case stage
      when 'after'
         scripts = @after_scripts
         dir = SCRIPT_DIR + '/after/'
      when 'before'
         scripts = @before_scripts
         dir = SCRIPT_DIR + '/before/'
      else
         raise ArgumentError.new('only after|before stage allowed')
      end

      scripts.each { |f|
         if @opts['profile']
            args = @opts['profile'].split('+')
            ret = system(@env_hash, dir + f, *args)
         else
            ret = system(@env_hash, dir + f)
         end

         case ret
         when true
            $log.info _(' "%{file}" script finished') % {:file => stage + '/' + f}
         when false
            $log.err _(' "%{file}" script failed: %{status}') % {:file => stage + '/' + f, :status => $?}
         when nil
            $log.err _(' "%{file}" execution fails') % {:file => stage + '/' + f}
         end
      }
   end

end # class Storage
