#
# = conf.rb: модуль для загрузки конфигурации (avreg.conf + командная строка)
#

require 'etc'
require 'optparse'

require 'avreg/unlink/unlink-conf-struct'
require 'avreg/unlink/confile'

module AVRegUnlinkConf

   class AVRegUnlinkConfError < RuntimeError
      # Reason which caused the error.
      Reason = 'parse error'.freeze

      def initialize(sym_or_fixnum, *args)
         @sym_or_fixnum = sym_or_fixnum
         @args = args
         @reason = nil
      end

      attr_reader :args
      attr_writer :reason

      #
      # Returns error reason. Override this for I18N.
      #
      def reason
         @reason || self.class::Reason
      end

      def inspect
         "#<#{self.class.to_s}: #{sym_or_fixnum} #{args.join(' ')}>"
      end

      #
      # Default stringizing method to emit standard error message.
      #
      def message
         reason + ': ' + "\"#{@sym_or_fixnum}\""
      end

      alias to_s message
   end

   class ReadOnlyError < AVRegUnlinkConfError
      const_set(:Reason, 'read only variable'.freeze)
   end

   class AVRegInvalidProfileConf < StandardError
   end

   include AVRegUnlinkConfStruct

   # FIXME we should not use grouped regexp in optparse
   SPACES_SIZE_REGX = /\A(\d+(?:\.\d+)*[KMGT]?|\d+(?:\.\d+)*%|`\s*.+\s*`\s*[KMGT]?)\Z/i
   UINT_REGX        = /\A\d+\Z/
   FLOAT_REGX       = /\A\d+(\.\d+)*\Z/
   SIZE_REGX_G      = /\A(\d+(\.\d+)*)([KMGT]?)\Z/i
   PROFILES_REGX    = /^[A-Za-z0-9][A-Za-z0-9_\.:\-]+$/

   def get_size_with_modifier(def_modifier, human_size)
      m = SIZE_REGX_G.match(human_size)
      raise OptionParser::InvalidArgument if m.nil? or m.length != 4
      size     = m[1]
      afterDot = m[2]
      modifier = m[3]
      modifier = def_modifier if modifier.nil? or modifier.empty?

      begin
        if afterDot.nil? or afterDot.empty?
          sizeNum = size.to_i
        else
          sizeNum = size.to_f
        end
      rescue
        raise OptionParser::InvalidArgument
      end

      k = case modifier
      when 'M','m' then 1000000
      when 'G','g' then 1000000000
      when 'T','t' then 1000000000000
      when 'K','k' then 1000
      else
         raise OptionParser::InvalidArgument
      end

      sizeNum *= k
      ret = sizeNum.to_i
   end # get_size_with_modifier

   # Массив для хранения конфигураций всех профилей avreg-unlink
   class AllProfilesOpts < Array

      def <<(profile_conf)
         newprof_name = profile_conf['profile']
         newprof_db = profile_conf.get_with_defaults('db-name')
         newprof_spool = profile_conf.get_with_defaults('storage-dir')
         newprof_cameras = profile_conf['cameras']

         diff = 0
         self.each do | prof |
            prof_name = prof['profile']
            prof_db = prof.get_with_defaults('db-name')
            prof_spool = prof.get_with_defaults('storage-dir')
            prof_cameras = prof['cameras']

            if newprof_db == prof_db and newprof_spool == prof_spool
               if prof_cameras.nil? or newprof_cameras.nil?
                  raise AVRegInvalidProfileConf,
                     _("The list of cameras (\"devlist\") for \"%{profile1}\" and/or \"%{profile2}\" profile is not defined") %
                     { :profile1 => newprof_name, :profile2 => prof_name }
               end

               # merge devlist, name = pname + pname, "+" as sign of merging XXX
               prof['profile'] += '+' + newprof_name
               prof['cameras'] = prof_cameras | newprof_cameras

               next # not append
            end

            # вот в этом месте уже что-то различается: db, spool или оба

            if newprof_db == prof_db and newprof_spool != prof_spool
               # база одна, но каталоги разные,
               # это допустимо лишь если заданы непересекающиеся devlist (cameras)
               if newprof_cameras.nil? or prof_cameras.nil?
                  raise AVRegInvalidProfileConf,
                     _("The list of cameras (\"devlist\") for \"%{profile1}\" and/or \"%{profile2}\" profile is not defined") %
                     { :profile1 => newprof_name, :profile2 => prof_name }
               end
               camdiff = newprof_cameras & prof_cameras
               if camdiff.length > 0
                  raise AVRegInvalidProfileConf,
                     _("The lists of cameras (\"devlist\") of the \"%{profile1}\" and \"%{profile2}\" profiles must not overlap") %
                     { :profile1 => newprof_name, :profile2 => prof_name }
               end
            end
            if newprof_db != prof_db and prof_spool == newprof_spool
               # разные базы, но один каталог
               raise AVRegInvalidProfileConf,
                  _("The \"storage-dir\" value of profile \"%{profile1}\" must not be identical with those of the profile \"%{profile2}\"") %
                  { :profile1 => newprof_name, :profile2 => prof_name }
            end

            diff += 1
         end

         # Добавляем новую конфигурацию профиля как независимую
         # только если были различия со _всеми_ пред. конфигурациями профилей
         # Если были совпадения хотя бы с одной пред. конф., то считаем такую конф. поглощённой.
         if diff == self.length
            super(profile_conf)
         end
      end

      def initialize(argv)
         super()

         sinlge_conf = UnlinkConfStruct.new

         off2 = '  '
         opt_parser = OptionParser.new(nil, 27, ' ' * 4) do |op|

            op.banner =  _("Usage: %{filename} [options] [CAMERAS_LIST] [\"FROM..TO\"]\n") % { :filename => $pr_name }
            op.banner << _("  where: CAMERAS_LIST limit set of cameras, i.e. \"1,2,7-13\"\n")
            op.banner << _("         \"FROM..TO\" set a time range, in SQL timestamp format \"YYYY-MM-DD[ hh:mm[:ss]]\"\n")

            op.banner << "\n"

            op.banner << _("Example: %{filename} -v 1-3,5-10\n") % { :filename => $pr_name }
            op.banner << "         #{$pr_name} -v 1-3,5-10 '2014-05-12 13:57..2014-07-19'\n"
            op.banner << "         #{$pr_name} -v 1-3,5-10 \"..2014-07-19 15:59\"\n"
            op.banner << "         #{$pr_name} -v 1-3,5-10 \"2014-05-12..\"\n"
            op.banner << "         #{$pr_name} -v --no-del --repair --optimize\n"

            op.separator ''
            op.separator _('  Common options:')

            op.on('-c', '--conf=PATH', String,
                          _('config file path'),
                          off2 + _('by default') + ": `#{DEF_OPTS['cnf']}'") do |path|
               sinlge_conf['cnf'] = path
                          end

            op.on('-p', '--profile=NAME', String,
                          _("profiles is additional configuration files:"),
                          off2 + _('see') + " `ls /etc/avred/profiles/[A-Za-z0-9]*',",
                          off2 + _('by default') + ': <' + _('empty') + '>') do |file|
               sinlge_conf['profile'] = file
                          end

            op.on('-u', '--user=NAME', String,
                          _("system user name for seteuid(2)"),
                          off2 + _('by default') + ': <' + _('empty') + '> - ' + _('no use') + ' seteuid(2)') do |user|
               sinlge_conf['uid']  = Etc::getpwnam(user).uid
               sinlge_conf['user'] = user
                          end
            op.on('-g', '--group=NAME', String,
                          _("system group name for setegid(2)"),
                          off2 + _('by default') + ': <' + _('empty') + '> - ' + _('no use') + ' setegid(2)') do |group|
               sinlge_conf['gid']   = Etc::getgrnam(group).gid
               sinlge_conf['group'] = group
                          end

            op.separator ""
            op.separator _('  Filesystem removal specific options:')

            op.on('--[no-]del',
                          _('actually delete the files or not') + ', ' +
                          _('by default') + ': ' + _('Yes')) do |nd|
               sinlge_conf['del'] = nd
                          end

            op.on('--[no-]hooks',
                          _('enable or disable') + ' "/etc/avreg/scripts/unlink/{before,after}/*"' + ', ',
                          off2 + _('hook scripts execution') + ', ' +
                          _('by default') + ': ' + _('Yes') + ', ' + ('enable')) do |nd|
               sinlge_conf['hooks'] = nd
                          end

            op.on('-k', '--disk-usage-cmd=CMD', String,
                          _("command to get disk space usage in 1KB blocks") + ',',
                          off2 + _("required ouput format: \"USED TOTAL\"") + ',',
                          off2 + _('by default') + ': <' + _('empty') + '>, ' + 'use statfs(2)') do |cmd|
               sinlge_conf['disk-usage-cmd'] = cmd
                          end

            op.on('-a', '--storage-dir=DIR', String,
                          _("media files' root path") + ',',
                          off2 + _('by default') + ": `#{DEF_OPTS['storage-dir']}'") do |path|
               sinlge_conf['storage-dir'] = path
                          end
            op.on('-m', '--removed-manually=FILE', String,
                          _('list of the manually removed media files') + ',',
                          off2 + _('by default') + ": `#{DEF_OPTS['removed-manually']}'") do |file|
               sinlge_conf['removed-manually'] = file
                          end

            # adding on the fly
            #op.on("--add=SWITCH=[ARG]", "add option on the fly", /\A(\w+)(?:=.+)?\Z/) do
            #   |opt, var|
            #   op.on("--#{opt}", "added in runtime", &eval("proc {|@#{var}|}"))
            #end

            op.separator ""
            op.separator _('         For the three following options you can specify')
            op.separator _('      k[ilobytes], M[egabytes] and G[igabytes] suffix modifiers')
            op.separator _('      for example: -s 2500M or -e 50G')
            op.separator _('         For "-space" options you can specify percent suffix,')
            op.separator _('      of total "storage-dir" device space, for example: -s\'10%\' -e\'20%\'')

            op.on('-s', '--start-space=SIZE', SPACES_SIZE_REGX,
                          _('min. free space of storage device when start cleaning'),
                          off2 + _('by default') + ': ' + _('space') + ' - "' +
                          DEF_OPTS['start-space'].to_s +  '", ' + _('modifier') + ' - "G"') {
               |size_def|  sinlge_conf['start-space'] = size_def.kind_of?(Array) ? size_def[1] : size_def
            }

            op.on('-e', '--end-space=SIZE', SPACES_SIZE_REGX,
                            _('min. free space of storage device when stop cleaning') + ',',
                          off2 + _('by default') + ': ' + _('space') + ' - "' +
                          DEF_OPTS['end-space'].to_s +  '", ' + _('modifier') + ' - "G"') {
               |size_def|  sinlge_conf['end-space'] = size_def.kind_of?(Array) ? size_def[1] : size_def
            }

            op.on('-d', '--del-per-once=SIZE', SIZE_REGX_G,
                          _('delete this SIZE volume per once') + ',',
                          off2 + _('by default') + ': ' + _('volume') + ' - ' +
                          DEF_OPTS['del-per-once'].to_hfs + ', ' +
                          _('modifier') + ' - "M"'   ) do |size|
                            sinlge_conf['del-per-once'] = get_size_with_modifier('M', size.kind_of?(Array) ? size[0] : size)
                            if sinlge_conf['del-per-once'] < 1000000
                              sinlge_conf['del-per-once'] = 1000000
                            end
                          end

            op.on('-t', '--timeout=SEC', FLOAT_REGX,
                          _('timeout in seconds for sleep middle deleting transactions') + ',',
                          off2 + _('by default') + ': ' + DEF_OPTS['timeout'].to_s + ' sec.' +
                          ', ' + _('allowable range of values') + ': [0.1-1.0].'
                 ) do |t|
                    f = case t
                        when Array
                           Float(t[0])
                        when String
                           Float(t)
                        else
                           raise OptionParser::InvalidArgument, _("\"%{_val}\", unsupported \"%{_par}\" value") %
                              { :_val => t, :_par => "timeout" }
                        end
                    if f < 0.1
                       f = 0.1
                    elsif f > 1.0
                       f = 1.0
                    end
                    sinlge_conf['timeout'] = f
                 end

            op.separator ""
            op.separator _("  Database specific options:")
            op.on("-L", "--[no-]lock",
                          _('try lock tables') + ',',
                          off2 + _('by default') + ': ' + _('No')) do |l|
               sinlge_conf['lock'] = l
                          end

            op.on('-R', '--[no-]repair',
                          _('repair all tables on start') + ',',
                          off2 + _('by default') +  ': ' + _('No')) do |r|
               sinlge_conf['repair'] = r
                          end

            op.on('-O', '--[no-]optimize',
                          _('optimize all tables before exit') + ',',
                          off2 + _('by default') + ': ' + _('No')) do |o|
               sinlge_conf['optimize'] = o
                          end

            op.on('-r','--rows-per-select=NUM', UINT_REGX,
                          _("select a NUM rows from the database in one sql query") + ',',
                          off2 + _('by default') + ': ' + DEF_OPTS['rows-per-select'].to_s) do |r|
               rps = r.to_i
               rps = 1000 if ( rps < 1000 )
               sinlge_conf['rows-per-select'] = rps
                          end

            op.on('-i','--select-now-minus=MIN', UINT_REGX,
                          _("use for `select ... where DT1 < (NOW() - MIN*60)' query") + ',',
                          off2 + _('by default') + ': ' + DEF_OPTS['select-now-minus'].to_s) do |r|
               snm = r.to_i
               snm = 1 if ( snm < 1 )
               snm = 24*60 if ( snm > 24*60 )
               sinlge_conf['select-now-minus'] = snm
                          end

            op.separator _("    (deprecated, you should set these parameters in the configuration file)")

            op.on('-T', '--db-type=TYPE', String,
                          _('sql client type') + ', ' + _('by default') + ": \"#{DEF_OPTS['db-type']}\"") do |sqltype|
               sinlge_conf['db-type'] = sqltype;
            end
            op.on('-H', '--db-host=HOSTNAME', String,
                          _('sql server hostname') + ', ' + _('by default') +
                          ': <' + _('empty') + '>, '+_('i.e.') +' local (unix socket)') do |host|
                  sinlge_conf['db-host'] = host
                          end
            op.on('--db-port=PORT', UINT_REGX,
                          _('sql server port') + ', ' + _('by default') + ': <' + _('empty') + '>') do |port|
               sinlge_conf['db-port'] = port.to_i
                          end
            op.on('-N', '--db-name=DATABASE', String,
                          _('sql database name') + ', ' + _('by default') + ": \"#{DEF_OPTS['db-name']}\"") do |name|
               sinlge_conf['db-name'] = name
                          end
            op.on('-U', '--db-user=USERNAME', String,
                          _('database user name') + ',',
                          off2 + _('by default') + ": \"#{DEF_OPTS['db-user']}\"") do |user|
               sinlge_conf['db-user'] = user
                          end
            op.on('-P', '--db-passwd=PASSWORD', String,
                          _('user password') + ', ' + _('by default') + ': <' + _('empty') + '>') do |passwd|
               sinlge_conf['db-passwd'] = passwd
                          end

            op.separator ""
            op.separator _("  Other options:")

            # boolean switch with optional argument(default false)
            verbose_regx   = Regexp.compile('\A(?:\d?|v+)\Z')
            verbose_regx_g = Regexp.compile('\A(\d?|v+)\Z')
            op.on("-v", "--verbose[=LEVEL]", verbose_regx, _("run verbosely, for example: -vvvv or -v3")) do |v|
               if v.nil? then
                  $verbose = 1
               else
                  vopt = verbose_regx_g.match(v)[1]
                  begin
                     $verbose = Integer(vopt)
                  rescue ArgumentError
                     $verbose = vopt.count('v') + 1
                  end
               end
            end

            # No argument, shows at tail.  This will print an sinlge_conf summary.
            # Try it and see!
            op.on_tail("-h", "--help", _("show this message")) do
               puts op
               exit 0
            end

            # Another typical switch to print the version.
            op.on_tail("-V", "--version", _("show version")) do
               puts $pr_name + '-' + $version
               exit 0
            end
         end

         begin
            opt_parser.parse!(argv)
            # for non --key parameters
            argv.each { |v|
               if v.include?('..')
                  sinlge_conf.parse_time_range(v)
               else
                  sinlge_conf.parse_cameras_range(v)
               end
            }
         rescue OptionParser::ParseError => err
            $log.err err.to_s
            # $stderr.puts "", opt_parser.help
            exit 1
         end

         # puts "cmdline has been parsed succesfully"
         # here cmdline options only in sinlge_conf
         # запрещаем установку значений опций, явно заданных в коммандной строке
         sinlge_conf.lock_assigned

         # читаем основной файл-конфиг
         main_cnf = sinlge_conf['cnf'] ? sinlge_conf['cnf'] : DEF_OPTS['cnf']
         if ( main_cnf )
            ConfFile.parse(opt_parser, sinlge_conf, main_cnf , $pr_name)
         end
         signle_conf_save = sinlge_conf.dup

         profiles_names = Array.new
         if sinlge_conf.profile
            # имя профиля явно указано в командной строке
            profiles_names << sinlge_conf.profile
         else
            Dir.glob('/etc/avreg/profiles/[A-Za-z0-9]*').each do | fpath |
               fname = File.basename(fpath)
               if File.file?(fpath) and PROFILES_REGX =~ fname
                  profiles_names << fname
               end
            end
            profiles_names.sort!
         end

         if sinlge_conf.is_readonly?('cameras')
            unless profiles_names.empty?
               # TODO: check profile[cameras] into cmdline[cameras] and use profiles
               $log.notice _("Ignore /etc/avreg/profiles/* configurations by cmdline \"cameras\" option")
            end
         else
            profiles_names.each do | profile_name |
               begin
                  $log.debug "Trying to load \"#{profile_name}\" profile configuration"
                  sinlge_conf = signle_conf_save.dup
                  sinlge_conf['profile'] = profile_name unless sinlge_conf['profile'] # may be frozen by cmdline setting
                  ConfFile.parse(opt_parser, sinlge_conf, "/etc/avreg/profiles/#{profile_name}" , $pr_name )
                  self << sinlge_conf
               rescue => detail
                  if profiles_names.length > 1
                     $log.err(detail.to_s + ' (' + detail.class.to_s + ')')
                     $log.err _("Ignore \"%{name}\" profile configuration") % { :name => profile_name }
                  else
                     raise # only signle profile - nothing to be removed
                  end
               end
            end
         end

         self << signle_conf_save.dup if self.empty? # no profiles

         if (
               self.count == 1                     and # only single or merged profiles
               not self[0]['cameras'].nil?         and # devlist is assigned
               not self[0].is_readonly?('cameras') and # devlist is not assigned by command line
               not self[0]['profile'].nil?         and # use profile(s) conf
               self[0]['profile'].include? ?+       # all (self.count == 1) profiles was merged (pname1+pname2+...)
            )
            # if devlist set by merged profiles only => speedup sql
            $log.debug("Reset devlist(cameras) value #{self[0]['cameras']} to speedup SQL");
            self[0]['cameras'] = nil
         end

         if ( (self.count > 1) and ( self[0]['to-date'] or self[0]['from-date'] ) )
            raise OptionParser::ParseError.new _("you must explicitly choose a single profile (see `-p' switch) when manually removing (FROM..TO)")
         end

         # пост-обработа некотых параметров, значения которых могут вычисляються
         # в зависимости от значений других параметров
         each do | __opts |
            __opts.final
            __opts.merge_defaults
            __opts.freeze
         end
         freeze
      end  # initialize

   end  # class AllProfilesOpts

end # module
