# encoding: UTF-8

=begin rdoc

Модуль описыват класс AVReg_DB для взаимодействия с базой данных:
 - чтения событий в таблице EVENTS,
 - удаление устаревших событий из таблицы EVENTS,
 - сервисные операции - оптимизация и удаление.
=end

require 'sequel'
require 'logger'

AVREG_SQL_DATETIME_FMT='%Y-%m-%d %H:%M:%S'

module TO_SQL_TIME
   def to_sql
      strftime(AVREG_SQL_DATETIME_FMT)
   end
end

# новое DBI (старше чем в lenny) отдаёт нативные типы вместо DBI::Timestamp, и т.п.
class Time
   include TO_SQL_TIME
end
class Date
   include TO_SQL_TIME
end
class DateTime
   include TO_SQL_TIME
   def to_time
      Time.parse(self.strftime('%FT%T%'))
   end
end

class Sequel::Error
      alias_method :orig_to_s, :to_s
      def to_s
         e1 = self
         .orig_to_s
         .dup
         .lines
         .first
         .strip

         #if e1.encoding == 'ASCII-8BIT' #  e1 =~ /^PG::/i
            e1 = e1.force_encoding(Encoding.default_external)
         #end

         return e1
      end
   end

#
# Класс описывающий доступ к базе данных.
#
class AVReg_DB

   MIN_ROWS_Q = 3.0/4.0

   # формат select запроса
   SELECT_SQL_FMT = <<_EOL_
select DT1 as dt1, DT2 as dt2,
CAM_NR as cam_nr, EVT_ID as evt_id,
FILESZ_KB as filesz_kb, FRAMES as frames,
EVT_CONT as evt_cont
from EVENTS
%s
order by DT1
limit %d
_EOL_

   # строка идентифицируемая базу данных используемая в DBI.connect 
   attr_reader :db_string

   # сколько записей в таблице EVENTS обработано программой за всё время работы программы 
   attr_reader :fetched

   # конструктор, подключаемся к базе данных
   def initialize(conf)
      # connect to the SQL server
      db_host   = conf['db-host']
      if db_host.nil? or db_host.empty?
         @db_string = conf['db-name']
      else
         @db_string = conf['db-name'] + ':' + db_host
      end
      case conf['db-type']
         when /mysql/i
            db_adapter = 'mysql2'
            @events_t = :EVENTS
         when /pgsql/i
            db_adapter = 'postgres'
            @events_t = :events
         else
            raise ArgumentError.new( _('unsupported SQL type "%{sqltype}"') %  { :sqltype => conf['db-type'] } )
      end

      $log.info _('Try connect to "%{sqltype}" database "%{db_string}"') % {
         :sqltype => conf['db-type'],
         :db_string => @db_string
      };

      @dbh = Sequel.connect(
         :adapter => db_adapter,
         :host => conf['db-host'],
         :port => conf['db-port'],
         :database => conf['db-name'],
         :user => conf['db-user'],
         :password => conf['db-passwd'],
         :logger => $verbose >= 2 ? Logger.new('/tmp/avreg-unlink.sql.log') : nil
      );

      dataset = @dbh['select CURRENT_TIMESTAMP as time, version() as ver'].map([:time, :ver]);
      row = dataset[0]
      server_time = row[0].to_time
      if $verbose >= 2 then
         # get server version string and display it
         $log.debug "  use #{db_adapter} adapter, server version: \"#{row[1]}\"; time: #{row[0]}"
         #$log.debug server_time.to_sql
      end

      @limit = conf['rows-per-select']
      if conf['from-date']
         @where_from_date = "DT1 >= '#{conf['from-date']}'"
      end
      if conf['to-date']
         @where_to_date = "DT1 <= '#{conf['to-date']}'"
      else
         # чтобы не было зацикливания, ограничиваем по NOW на сервере SQL
         # а select-now-minus нужен чтобы в конце работы не удалить только что записанные файлы
         @where_to_date = "DT1 < '#{(server_time - (conf['select-now-minus'] * 60)).to_sql}'"
      end
      if conf['cameras']
         @where_cameras = "CAM_NR in (#{conf['cameras'].join(',')})"
      end

      @where_cond = ''
      class << @where_cond
         def and(sql_expr)
            if empty?
               self << 'where '
            else
               self << ' and '
            end
            self << sql_expr
         end
         def or
            if empty?
               self << 'where '
            else
               self << ' or '
            end
         end

         def reset
            unless empty?
               if respond_to? :clear
                  clear
               else
                  gsub!(/./, '')
               end
            end
            self
         end
      end

      @latest_fetched_dt1 = nil
      @fetched = 0
   rescue => e
      $log.crit _('connect failed') + ': ' + e.to_s
      raise
   end # initialize

   # отключаемся от базы данных
   def close!
      if not @dbh.nil? then
         $log.debug _('Disconnect from database "%{db_string}"') % {:db_string => @db_string }
         @dbh.disconnect
         remove_instance_variable(:@dbh)
      end
   end # close!

   # выполняется OPTIMIZE sql для MySQL 
   # FIXME или VACUUM в перспективе для других серверов
   def optimize(cmd='optimize')
      unless @dbh.opts[:adapter]=~ /mysql/i
         $log.notice _("skip %{cmd} table for %{adapter} sql-server") % {
            :cmd => cmd,
            :adapter => @dbh.opts[:adapter]
         }
         return 0
      end
      errcnt=0
      tbllist = @dbh.tables.map {|t| t.to_s}.join(', ')
      sql = cmd + ' tables ' + tbllist;
      $log.notice "sql: `#{sql}'"
      @dbh[sql].each do |row|
         if row[:Msg_type] === 'status'
            $log.info(sprintf("  %s   %-18s  %-6s = `%s'\n", row[:Op], row[:Table], row[:Msg_type], row[:Msg_text]))
         else
            errcnt += 1
            $log.crit(sprintf("\a  %s   %-18s  %-6s = `%s'\n", row[:Op], row[:Table], row[:Msg_type], row[:Msg_text]))
         end
      end
      return errcnt
   rescue => e
      $log.crit "#{cmd} SQL failed: #{e}"
      raise
   end # repair

   # выполняется REPAIR для MySQL
   def repair
      self.optimize('repair')
   end # optimize

   # читает очередную "старую" порцию записей в таблице EVENTS
   # и возвращает (итератор) по одной строке (yuild row)
   def select_olderst
      raise 'You must use block select_olderst {|row| ... }' unless block_given?
      done = false

      until done do
         @where_cond.reset
         if @latest_fetched_dt1
            @where_cond.and "DT1 > '#{@latest_fetched_dt1.to_sql}'"
         else
            @where_cond.and "#@where_from_date" if @where_from_date
         end
         @where_cond.and "#@where_to_date" if @where_to_date
         @where_cond.and "#@where_cameras" if @where_cameras

         rows = @dbh[sprintf(SELECT_SQL_FMT, @where_cond, @limit)].all
         selected = rows.count
         $log.debug "  sql: select ... #@where_cond order by DT1 limit #@limit => #{selected}"
         break if rows.empty?

         if selected == @limit # выбрали с ограничением, поэтому нужно избавиться от последней сек.
            lastest_row = rows.last
            while row = rows.last
               break if row[:dt1] < lastest_row[:dt1]
               rows.pop
            end
            if rows.empty? or ( (rows.count.to_f / selected) < MIN_ROWS_Q )
               last_limit = @limit
               @limit *= 2
               $log.warning "increase select limit (rows-per-select) from #{last_limit} to #@limit [#{rows.count}]"
               redo
            end
         end

         rows.each do |r|
            if done
               # дочитываем последнюю секунду полностью
               break if r[:dt1] > @latest_fetched_dt1
               @fetched += 1
               yield r
            else
               @fetched += 1
               @latest_fetched_dt1 = r[:dt1]
               done = yield r
            end
         end

         break if selected < @limit # не осталось записей с учётом макс. диапазона
      end
   end # select_olderst

   # удаляет обработанные записи
   def del_fetched!
      raise 'Unknown @latest_fetched_dt1' unless @latest_fetched_dt1
      return 0 if @fetched == 0

      to_delete_events = @dbh[@events_t];
      if @where_from_date
         to_delete_events = to_delete_events.where(Sequel.lit(@where_from_date))
      end
      to_delete_events = to_delete_events.where(Sequel.lit("DT1 <= '#{@latest_fetched_dt1.to_sql}'"))
      if @where_cameras
         to_delete_events = to_delete_events.where(Sequel.lit(@where_cameras))
      end
      $log.debug "  sql: #{to_delete_events.delete_sql}"
      to_delete_events.delete
   rescue => e
      $log.crit "`delete from EVENTS ...' SQL failed: #{e}"
      return 0
   end # del_fetched!

end # class AVReg_DB
