#
# = fs.rb: модуль с классами для взаимодействия с файловой системой
#
# Удаление старых файлов из архива и рекурсивное удаление пустых каталогов. 
#

require 'avreg/utils'

# Рекурсивное удаление непустых каталогов
class RmDirsRecur
   def self.rmdirs!(dirs,prefix)
      glob_rmdir_hit = 0
      return 0 if dirs.nil? or dirs.empty?
      dir_level_rsort!(dirs)
      dirs.uniq!

      loop do
         rmdir_hit = 0
         cnt = dirs.length
         break if cnt == 0
         
         for i in 0..(cnt-1)
            dir = dirs[i]
            begin
               Dir.rmdir(prefix + '/' + dir)
               own_dir = dir.gsub(/\A(.+)\/[^\/]*/,'\1')
               # p "Owner dir is [ #{own_dir} ]"
               dirs[i] = own_dir
               rmdir_hit += 1
               puts "rmdir `#{dir}'" if $verbose > 2
            rescue Errno::ENOTEMPTY
               next
            rescue SystemCallError
               # удаляем каталог из списка
               # $stderr.puts $!
               dirs[i] = nil
            end
         end

         break if rmdir_hit == 0
         dirs.compact!
         dirs.uniq!
         dir_level_rsort!(dirs)
         glob_rmdir_hit += rmdir_hit
      end

      if $verbose > 2
         dirs.compact!
         dirs.uniq!
         if not dirs.empty?
            puts _("The %{nbr} dirs was stay undeleted:") % { :nbr => dirs.length }
            pp dirs
         end
      end
      glob_rmdir_hit
   end # rmdirs!

   def self.dir_level_rsort(ar)
      ar.sort { |x,y|
         y.split('/').length <=> x.split('/').length
      }
   end
   
   def self.dir_level_rsort!(ar)
      ar.sort! { |x,y|
         y.split('/').length <=> x.split('/').length
      }
   end

end # class RmDirsRecur

# 
# Удаляет устаревшие файлы согласно переданного списка
# с контролем свободного места на устройстве.
#
class SorageFS
   attr_reader :start_free_space, :broken_links, :deleted_files, :del_size, :fetched_files

   def initialize(opts)
      @opts = opts

      @root                 = @opts['storage-dir']
      @need_free_space      = @opts['end-space']
      @del_per_once         = @opts['del-per-once']
      @del_per_once_timeout = @opts['timeout']
      @broken_links_file    = @root + '/' +  @opts['removed-manually'] if @opts['removed-manually']

      # statictics
      @del_size      = 0
      @last_del_size = 0
      @fetched_files = 0
      @deleted_files = 0
      @broken_links_f = nil
      @broken_links = 0
      @removed_dirs = 0

      # dir
      @dirs = []

      begin
         load_fs_sizes!
         @start_free_space = @fs.avail
         @broken_links_f = File.new(@broken_links_file, "a", 0644) if @broken_links_file and @opts['del']
      rescue
         $log.crit "SorageFS : #{$!.message}"
         raise
      end
   end # initialize

   def close!
      if @broken_links_f
         if ( @broken_links > 0 ) 
            @broken_links_f.puts   ';--------------------------------------------------------'
            @broken_links_f.puts _("; found %{br} hands-removed files") % { :br => @broken_links}
            @broken_links_f.puts   ';--------------------------------------------------------'
            @broken_links_f.puts
         end
         @broken_links_f.close
         @broken_links_f = nil
      end
   end

   def done?
      curdel_size = @del_size - @last_del_size # будет =0 когда удалили @del_per_once
      res = ( @fs.avail + curdel_size ) > @need_free_space
      #puts "( #{@fs.avail.to_hfs} + #{curdel_size.to_hfs} ) > #{@need_free_space.to_hfs} ==> #{res}"
      res
   end # done?

   # return true if done?
   def unlink!(row)
      dt1, dt2, cam_nr, evt, filesize_KiB, frames, evt_content = row.values

      return done? if evt_content.nil? or evt_content.empty?
      evt_id = Integer(evt)
      return done? unless ( evt_id == 12 or evt_id == 23 or evt_id == 32 or ( evt_id >= 15 and evt_id <= 21 ) )

      evt_content.strip!
      fullname = @root + '/' + evt_content
      @fetched_files += 1

      begin
         File.unlink(fullname)
         @del_size += filesize_KiB * 1024
         @deleted_files += 1
      rescue Errno::ENOENT
         if $verbose >= 2
            $log.warning "#{fullname} not exist"
         end

         @broken_links += 1
         if @broken_links_f
            begin
               if ( @broken_links == 0 )
                  @broken_links_f.puts ';========================================================'
                  @broken_links_f.puts '; ' + Time.now.ctime
                  @broken_links_f.puts ';--------------------------------------------------------'
               end
               @broken_links_f.printf("cam #%2d, %-30s, %9s, %s\n",
                                      cam_nr,
                                      if dt1.eql?(dt2)
                                         dt1.to_sql
               else
                  dt1.to_sql + ' [ ' + (dt1.to_time.to_i - dt2.to_time.to_i).to_duration + ' ]'
               end,
               (filesize_KiB*1024).to_hfs, evt_content )
            rescue => err
               $log.err "writing to \"#{@broken_links_file}\" failed -> #{err}"
               @broken_links_f = nil
            end
         end
      rescue ArgumentError
         $log.err _("database table EVENTS might be corrupted, row dump: %s") % row.inspect
         raise
      end

      # save directory
      dir =  File.dirname(evt_content)
      @dirs.push(dir) unless @dirs.rindex(dir)

      unless $quit
         if ( @del_size - @last_del_size ) >= @del_per_once
            # on each @del_per_once space deletion
            load_fs_sizes!
            @last_del_size = @del_size
            if $verbose >= 1
               $log.notice _(" the total number of deleted files - %{num} at %{size}") %
               { :num => @deleted_files, :size => @del_size.to_hfs }
            end
            sleep(@del_per_once_timeout)
         end
      end

      done?
   end # unlink!

   def unlink_empty_dirs!
      @removed_dirs = RmDirsRecur.rmdirs!(@dirs, @root)
   end # unlink_empty_dirs!

   def PrintBeforeStat
      arr = Array.new
      arr.push( sprintf(_('real storage device is') + ' `%s\', ' + _('mount on') + ' `%s\', fs `%s\'' + "\n",
                        @fs.dev,
                        @fs.mntpoint,
                        @fs.fstype))
      arr += GetStat(_('before clean'))
   end # PrintBeforeStat

   def PrintAfterStat
      load_fs_sizes!

      arr = Array.new
      arr += GetStat(_('after clean '))
      arr.push( _("total deleted %{u1} files with common size %{s} and %{u2} dirs\n") %
                 { :u1 => @deleted_files,
                   :s  => @del_size.to_hfs,
                   :u2 => @removed_dirs } )
   end # PrintAfterStat

   def PrintWarn
     if @broken_links > 0
        $log.warning _("found %{broken} removed manually files, see %{s}") % { :broken => @broken_links, :s => @broken_links_file }
     end
   end

   private
   def GetStat(pref="")
      arr = Array.new
      arr.push(sprintf('%s `%s\' ' + _('space: total/usage/free') + ': %s / %s (%d%%) / %s (%d%%)' + "\n",
                       pref,
                       @fs.dev,
                       @fs.total.to_hfs,
                       @fs.used.to_hfs,  (@fs.used.to_f*100/@fs.total.to_f).round,
                       @fs.avail.to_hfs, (@fs.avail.to_f*100/@fs.total.to_f).round))
   end

   private
   def load_fs_sizes!
      @fs = StatFS.new(@root, @opts['disk-usage-cmd'])
   end


end # class StorageFS
