require 'avreg/utils'

# Переопределяем класс StatFS с учётом использования внешней команды
# для получения размеров блочных устройств вместо statfs(2)
# Это нужно при использовании дисковых квот для ограничения
# размеров архива AVReg.
module AVRegUtils
   class StatFS
      TWODIG_RGX = /\A(\d+)\s*(\d+)\Z/i

      alias :initialize_orig :initialize
      def initialize(path, disk_usage_cmd)
         Dir.chdir(path)
         if disk_usage_cmd
            @path  = path
            @bsize = 1024

            outline = %x{#{disk_usage_cmd}}.chomp
            unless $?.exitstatus.zero?
               raise _("unexpected \"disk-usage-cmd\" value: \"%s\"") % disk_usage_cmd
            end
            #puts "`#{disk_usage_cmd}` #{$?.exitstatus}, outline '#{outline}'"
            m = TWODIG_RGX.match(outline)
            einval(outline) unless m and m.size == 3

            realy_used = m[1].to_i
            limit      = m[2].to_i

            if realy_used < 0 or limit <= 0
            end

            realy_free = limit - realy_used
            if false and realy_free < 0 # перебор
               @blocks = realy_used
               @bfree  = @bavail  = 0
            else
               @blocks = limit
               @bfree  = @bavail = realy_free
            end

            calc_vars!
         else
            initialize_orig(path)
         end
         @m = Mtab.new.find(path)
      end

      public
      def dev
         @m[:spec]
      end

      def mntpoint
         @m[:file]
      end

      def fstype
         @m[:vfstype]
      end

      private
      def einval(outline)
         raise _("invalid \"disk-usage-cmd\" output value: \"%s\"") % outline
      end

   end
end

module AVRegUnlinkConfStruct
   DEF_SPACESIZE_MODIFIER = 'G'
   UNLINK_DATE_FMT        = '\d{4}\-[01]\d\-[0-3]\d(\s+[012]\d:[0-5]\d)?(:[0-5]\d)?'

   DEF_OPTS = {
      'cnf'              => '/etc/avreg/avreg.conf',
      'profile'          => nil,
      'user'             => 'avreg',
      'uid'              => Etc::getpwnam('avreg').uid,
      'group'            => 'avreg',
      'gid'              => Etc::getgrnam('avreg').gid,

      'del'              => true,
      'hooks'            => true,
      'disk-usage-cmd'   => nil,
      'storage-dir'      => '/var/spool/avreg',
      'removed-manually' => 'removed-manually.txt',
      'start-space'      => '10%',
      'end-space'        => '15%',
      'del-per-once'     => 500000000,
      'timeout'          => 0.3,

      'from-date'        => nil,
      'to-date'          => nil,
      'cameras'          => nil,

      'lock'             => nil,
      'repair'           => nil,
      'optimize'         => nil,
      'rows-per-select'  => 5000,
      'select-now-minus' => 60,

      'db-type'          => 'Mysql',
      'db-host'          => nil,
      'db-port'          => nil,
      'db-name'          => 'avreg6_db',
      'db-user'          => 'avreg-unlink',
      'db-passwd'        => nil,
   }.freeze

   class UnlinkConfStruct < Struct.new('UnlinkConfStruct', *(DEF_OPTS.keys))

      def initialize
         @hack1 = false
         @ro_list = []
      end

      def key_normalize(key)
         case key
         when String
            name = key.gsub('_','-')
         when Symbol
            name = key.id2name.gsub('_','-')
         else
            raise "Invalid \"#{key}\" class #{key.class}"
         end
      end

      def key_denormalize(key)
         case key
         when String
            name = key.gsub('-','_')
         when Symbol
            name = key.id2name.gsub('-','_')
         else
            raise "Invalid \"#{key}\" class #{key.class}"
         end
      end

      alias :orig_getter :[]
      def [] key
         orig_getter key
      rescue NameError
         orig_getter key.to_s.gsub('_','-')
      end

      def is_readonly?(key)
         name = key_normalize(key)
         @ro_list.include?(name) and not @hack1
      end

      alias :orig_setter :[]=
      def []=(key, obj)
         # puts "self[#{key}] = #{obj}"
         name = key_normalize(key)
         raise ReadOnlyError.new(key, obj) if @ro_list.include?(name) and not @hack1
         orig_setter name, obj
      end

      def method_missing(method, *args)
         str = method.id2name.gsub('_','-')
         last_char_pos = str.length - 1
         # >=1.9 and <=1.8.7 string index styles
         if str[last_char_pos] == '=' or str[last_char_pos] == 61
            str = str[0..(last_char_pos-1)]
            # puts str
            self[str]= *args
         else
            self[str]
         end
      end

      def include?(key)
         self[key]
         true
      rescue
         false
      end
      alias :has_key? :include?
      alias :member?  :include?
      alias :key?     :include?

      def lock_assigned
         each_pair { |key, value| @ro_list << key_normalize(key) unless value.nil? }
      end

      def get_with_defaults(key)
         self[key] || DEF_OPTS["#{key}"]
      end

      def var_subst(str)
         return nil unless str
         str.gsub(/@[A-Z]+/) do |w|
            case  w[1..1000]
            when 'USER'     then get_with_defaults('user')
            when 'GROUP'    then get_with_defaults('group')
            when 'DIR'      then get_with_defaults('storage-dir')
            when 'FS'       then StatFS.new(get_with_defaults('storage-dir'), nil).dev
            when 'MNTPOINT' then StatFS.new(get_with_defaults('storage-dir'), nil).mntpoint
            else w
            end
         end
      end

      def parse_space(space_name, size_def)
         case size_def
         when /\A(\d+(?:\.\d+)*[tgmk]?)\Z/i
            get_size_with_modifier(DEF_SPACESIZE_MODIFIER, Regexp.last_match(1))
         when /\A(\d+(?:\.\d+)*)%\Z/
            fs = StatFS.new(get_with_defaults('storage-dir'), var_subst(get_with_defaults('disk-usage-cmd')))
            Integer((fs.total * Regexp.last_match(1).to_f) / 100.0)
         when /\A`\s*(.+)\s*`\s*([tgmk]?)\Z/i
            cmd      = Regexp.last_match(1).strip
            modifier = Regexp.last_match(2)
            modifier = '' if modifier.nil? or modifier.empty?
            raise if cmd.nil? or cmd.empty?
            cmd = var_subst(cmd)
            lim = %x{#{cmd}}.chomp
            puts "`#{cmd}` #{$?.exitstatus}, lim '#{lim}', mod '#{modifier}'"
            raise if $?.exitstatus != 0 or not lim
            parse_space(space_name, lim + modifier)
         else
            raise OptionParser::InvalidArgument.new "\"#{size_def}\" for \"#{space_name}\" option"
         end

      end

      def final
         @hack1 = true

         %w( start-space end-space ).each do | par |
            size_def = get_with_defaults(par)
            # pp "self[#{par}] = `#{size_def}'"
            self[par] = parse_space(get_with_defaults(par), size_def)
         end


         if (self['end-space'] - self['start-space'])  < 1000000
            msg = _('"end-space"(%{_end}) - "start-space"(%{_start}) mast be > 1MB') %
                  { :_end => self['end-space'].to_hfs, :_start => self['start-space'].to_hfs }
            if self['to-date'] or self['from-date']
               $log.warning msg
            else
               $log.err msg
               raise OptionParser::InvalidArgument.new 'start-space >= end-space'
            end
         end

         cmd = get_with_defaults('disk-usage-cmd')
         if cmd
            self['disk-usage-cmd'] = var_subst(cmd)
         end

         @hack1 = false
      end # parse_spaces

      def parse_cameras_range cams_def
         return if cams_def.nil?
         cams_array = Array.new
         if cams_def
            raise unless /\A([\d\.\-, ]+)\Z/.match(cams_def)
            cams_def.split(%r{,\s*}).each do |chunk|
               case chunk
               when /\A(\d+)(-|\.\.)(\d+)\Z/
                  r = Range.new(Integer($~[1]),Integer($~[3]))
                  cams_array.concat r.to_a
               when /\A\d+\Z/
                  cams_array << Integer(chunk)
               else
                  raise OptionParser::InvalidArgument.new(cams_def)
               end
            end
         end
         self['cameras'] = cams_array.sort!.uniq unless cams_array.empty?
      end # parse_manual_rm

      def parse_time_range date_def
         return if date_def.nil?
         raise unless /\A(#{UNLINK_DATE_FMT}|)\.\.(|#{UNLINK_DATE_FMT})\Z/.match(date_def)
         self['from-date'] = $~[1] if $~[1] and not $~[1].empty?
         self['to-date']   = $~[4] if $~[4] and not $~[4].empty?
      rescue
         raise OptionParser::InvalidArgument.new(date_def)
      end # parse_manual_rm

      def merge_defaults
         each_pair { |name, value| self[name] = DEF_OPTS["#{name}"] if value.nil? and not DEF_OPTS["#{name}"].nil? }
         self
      end

      def to_env
         env = Hash.new
         each_pair {|key, value|
            _key = key_denormalize(key)
            _val = case value
                   when Array
                      value.join(' ')
                   else
                      value.to_s
                   end
            env["AVREG_#{_key}"] = _val
         }
         env
      end
   end # class UnlinkConfStruct
end # module AVRegUnlinkConfStruct
